from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin

from wbportfolio.models import Adjustment
from wbportfolio.serializers import AdjustmentModelSerializer

from .configs import (
    AdjustmentButtonConfig,
    AdjustmentDisplayConfig,
    AdjustmentEndpointConfig,
    AdjustmentTitleConfig,
)


class AdjustmentModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = Adjustment.objects.select_related(
        "instrument",
        "last_handler",
    )
    serializer_class = AdjustmentModelSerializer

    filterset_fields = {
        "date": ["gte", "exact", "lte"],
        "instrument": ["exact"],
        "factor": ["gte", "exact", "lte"],
        "status": ["exact"],
        "last_handler": ["exact"],
    }
    ordering = ["-date"]
    ordering_fields = ["date", "instrument", "factor"]

    display_config_class = AdjustmentDisplayConfig
    title_config_class = AdjustmentTitleConfig
    endpoint_config_class = AdjustmentEndpointConfig
    button_config_class = AdjustmentButtonConfig


class AdjustmentEquityModelViewSet(AdjustmentModelViewSet):
    filterset_fields = {
        "date": ["gte", "exact", "lte"],
        "factor": ["gte", "exact", "lte"],
        "status": ["exact"],
        "last_handler": ["exact"],
    }

    def get_queryset(self):
        return super().get_queryset().filter(instrument=self.kwargs["instrument_id"])



from .trade_proposals import TradeProposalModelSerializer, ReadOnlyTradeProposalModelSerializer
from .claim import (
    ClaimAccountSerializer,
    ClaimAPIModelSerializer,
    ClaimCustomerModelSerializer,
    ClaimModelSerializer,
    ClaimRepresentationSerializer,
    ClaimTradeModelSerializer,
    NegativeTermimalAccountPerProductModelSerializer,
)
from .dividends import DividendModelSerializer, DividendRepresentationSerializer
from .expiry import ExpiryModelSerializer, ExpiryRepresentationSerializer
from .fees import FeesModelSerializer, FeesRepresentationSerializer
from .trades import (
    TradeModelSerializer,
    TradeProposalRepresentationSerializer,
    TradeRepresentationSerializer,
    TradeTradeProposalModelSerializer,
    ReadOnlyTradeTradeProposalModelSerializer,
)
from .transactions import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)



from wbcore.metadata.configs.titles import TitleViewConfig
from wbfdm.models import Instrument

from wbportfolio.models import Portfolio


class TradeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if self.request.GET.get("is_customer_trade", "False") == "True":
            return "Subscription / Redemption"
        return "Trade"

    def get_instance_title(self):
        return "Trade: {{shares}} {{bank}} - {{transaction_date}} ({{_product.name}})"

    def get_create_title(self):
        return "New Trade"


class TradeInstrumentTitleConfig(TradeTitleConfig):
    def get_list_title(self):
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Trades of Instrument {instrument.name} ({instrument.isin})"


class TradePortfolioTitleConfig(TradeTitleConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Trades within Portfolio {str(portfolio)})"


class CustodianDistributionInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Custodians Distribution {instrument.computed_str}"


class CustomerDistributionInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Customer Distribution {instrument.computed_str}"


class SubscriptionRedemptionTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            instrument = Instrument.objects.get(id=instrument_id)
            return f"Subscription / Redemption of {instrument.computed_str}"
        return "Subscription / Redemption"

    def get_instance_title(self):
        return "Subscription / Redemption: {{shares}} {{bank}} - {{transaction_date}} ({{_product.name}})"



from django.contrib import admin

from wbportfolio.models import AssetPosition


@admin.register(AssetPosition)
class AssetPositionModelAdmin(admin.ModelAdmin):
    list_filter = ("underlying_instrument", "portfolio")
    search_fields = (
        "underlying_instrument__name",
        "underlying_instrument__ticker",
    )
    # date_hierarchy = "date"
    list_filter = ("underlying_instrument__instrument_type",)
    list_display = (
        "date",
        "_price",
        "currency",
        "_currency_fx_rate",
        "_shares",
        "_total_value_fx_usd",
        "weighting",
        "portfolio",
        "underlying_instrument",
    )
    readonly_fields = [
        "_market_capitalization",
        "_price",
        "_currency_fx_rate",
        "_price_fx_portfolio",
        "_total_value",
        "_total_value_fx_usd",
        "_total_value_fx_portfolio",
        "_market_share",
    ]
    autocomplete_fields = [
        "portfolio",
        "currency",
        "underlying_instrument",
    ]
    ordering = ("-date",)
    raw_id_fields = [
        "import_source",
        "applied_adjustment",
        "underlying_quote_price",
        "currency_fx_rate_instrument_to_usd",
        "currency_fx_rate_portfolio_to_usd",
    ]



# Generated by Django 4.2.9 on 2024-02-16 08:33

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0055_remove_product__custom_management_rebates_and_more"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="companyportfoliodata",
                    name="assets_under_management_currency",
                ),
                migrations.RemoveField(
                    model_name="companyportfoliodata",
                    name="company",
                ),
                migrations.RemoveField(
                    model_name="companyportfoliodata",
                    name="potential_currency",
                ),
                migrations.RemoveField(
                    model_name="geographicfocus",
                    name="company",
                ),
                migrations.RemoveField(
                    model_name="geographicfocus",
                    name="country",
                ),
                migrations.DeleteModel(
                    name="AssetAllocation",
                ),
                migrations.DeleteModel(
                    name="AssetAllocationType",
                ),
                migrations.DeleteModel(
                    name="CompanyPortfolioData",
                ),
                migrations.DeleteModel(
                    name="GeographicFocus",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(name="AssetAllocation", table="company_portfolio_assetallocation"),
                migrations.AlterModelTable(name="AssetAllocationType", table="company_portfolio_assetallocationtype"),
                migrations.AlterModelTable(
                    name="CompanyPortfolioData", table="company_portfolio_companyportfoliodata"
                ),
                migrations.AlterModelTable(name="GeographicFocus", table="company_portfolio_geographicfocus"),
            ],
        )
    ]



import numpy as np
import pandas as pd

from .utils import _get_exchange_from_ticker, _get_ticker, file_name_parse_isin

COLUMN_MAP = {
    "Ticker": "underlying_instrument__ticker",
    "Underlying": "underlying_instrument__name",
    "Underlying ISIN": "underlying_instrument__isin",
    "Div Crncy": "currency__key",
    "Quantity": "shares",
    "Fx Rate": "currency_fx_rate",
    "Net Amount": "total_value_fx_portfolio",
    "Retro in%": "retrocession",
    "Gross Div": "price_gross",
    "Ex Div Date": "transaction_date",
    "Value Date": "value_date",
}


def parse(import_source):
    parts = file_name_parse_isin(import_source.file.name)
    product_data = parts["product"]

    df = pd.read_csv(import_source.file, sep=";", skipinitialspace=True)
    df = df.rename(columns=COLUMN_MAP).astype("str")

    df["underlying_instrument__exchange"] = df["underlying_instrument__ticker"].apply(
        lambda x: _get_exchange_from_ticker(x)
    )
    df["underlying_instrument__ticker"] = df["underlying_instrument__ticker"].apply(lambda x: _get_ticker(x))
    df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
    df["value_date"] = pd.to_datetime(df["value_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
    df = df[df["underlying_instrument__isin"].str.contains("([A-Z]{2})([A-Z0-9]{9})([0-9]{1})", regex=True)]
    float_columns = ["shares", "total_value_fx_portfolio", "price_gross", "currency_fx_rate", "retrocession"]
    for float_column in float_columns:
        df[float_column] = df[float_column].str.replace(" ", "").astype("float")
    df = df.drop(columns=df.columns.difference(COLUMN_MAP.values()))

    df["retrocession"] = df["retrocession"] / 100.0
    df["total_value"] = df["total_value_fx_portfolio"] / df["currency_fx_rate"]
    df["price"] = df["total_value"] / df["shares"] / df["retrocession"]
    df["total_value_gross"] = df["price_gross"] * df["shares"] * df["retrocession"]
    df["total_value_gross_fx_portfolio"] = df["total_value_gross"] * df["currency_fx_rate"]
    df["portfolio"] = [{"instrument_type": "product", **product_data}] * df.shape[0]
    df = df.replace([np.inf, -np.inf, np.nan], None)
    return {
        "data": df.to_dict("records"),
        "history": {
            "transaction_date": parts["valuation_date"].strftime("%Y-%m-%d"),
            "product": product_data,
        },
    }



import datetime
import re
from io import BytesIO

import pandas as pd

from wbportfolio.models import Product, Trade


def parse(import_source):
    df_dict = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name=None)

    data = list()
    product_isins = set()
    max_date = datetime.date(1900, 1, 1)

    for sheet_name, df in df_dict.items():
        if "prices" not in sheet_name:
            isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", sheet_name)
            isin = isin[0]
            product_isins.add(isin)
            df = df.rename(
                columns={
                    "Trade Date": "transaction_date",
                    "Price": "price",
                    "Nominal": "nominal",
                    "Market CP": "bank",
                    "Way": "way",
                }
            )
            df["transaction_date"] = pd.to_datetime(
                df["transaction_date"],
            )
            for trade in df.to_dict("records"):
                max_date = max(trade["transaction_date"].date(), max_date)

                nominal = trade["nominal"] if trade["way"] == "S" else trade["nominal"] * -1
                data.append(
                    {
                        "underlying_instrument": {"isin": isin, "instrument_type": "product"},
                        "portfolio": {"isin": isin, "instrument_type": "product"},
                        "transaction_date": trade["transaction_date"].strftime("%Y-%m-%d"),
                        "nominal": nominal,
                        "transaction_subtype": Trade.Type.REDEMPTION if nominal < 0 else Trade.Type.SUBSCRIPTION,
                        "bank": trade["bank"],
                        "price": round(trade["price"] / 10, 6),
                    }
                )
    underlying_instruments = Product.objects.filter(isin__in=product_isins).values_list("id", flat=True)
    return {
        "data": data,
        "history": {
            "underlying_instruments": list(underlying_instruments),
            "transaction_date": max_date.strftime("%Y-%m-%d"),
        },
    }



from .adjustments import AdjustmentButtonConfig
from .assets import (
    AssetPositionButtonConfig,
    AssetPositionInstrumentButtonConfig,
    AssetPositionPortfolioButtonConfig,
    DistributionChartButtonConfig,
    DistributionTableButtonConfig,
)
from .claims import ClaimTradeButtonConfig, ConsolidatedTradeSummaryButtonConfig

from .custodians import CustodianButtonConfig
from .fees import FeesButtonConfig
from .portfolios import PortfolioButtonConfig
from .products import ProductButtonConfig, ProductCustomerButtonConfig
from .registers import RegisterButtonConfig
from .trades import (
    TradeButtonConfig,
    TradeInstrumentButtonConfig,
)
from .trade_proposals import TradeProposalButtonConfig
from .reconciliations import AccountReconciliationButtonViewConfig, AccountReconciliationLineButtonViewConfig
from .signals import *



import factory

from wbportfolio.factories.portfolios import PortfolioFactory
from wbportfolio.models import PortfolioCashTarget


class PortfolioCashTargetFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PortfolioCashTarget

    valid_date = factory.Faker("date")
    portfolio = factory.SubFactory(PortfolioFactory)
    min_target = factory.Faker("pydecimal", left_digits=0, right_digits=4, max_value=0.01)
    target = factory.Faker("pydecimal", left_digits=0, right_digits=4, max_value=0.01)
    max_target = factory.Faker("pydecimal", left_digits=0, right_digits=4, min_value=0.01)
    comment = factory.Faker("paragraph", nb_sentences=1)



from wbcore.metadata.configs.titles import TitleViewConfig
from wbfdm.models import Instrument

from wbportfolio.models import Portfolio


class AssetPositionTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Asset Position"


class AssetPositionPortfolioTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Asset Positions for Portfolio: {str(portfolio)}"


class AssetPositionInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Implemented Portfolios for Instrument: {instrument.computed_str}"


class ContributorPortfolioChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Contributors: {str(portfolio)}"


class AssetPositionUnderlyingInstrumentChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Allocation charts"


class CashPositionPortfolioTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Cash Positions per Portfolio"


class DistributionChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Distribution Chart - {str(portfolio)}"


class DistributionTableTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        return f"Distribution Table - {str(portfolio)}"


class CompositionModelPortfolioPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        portfolio = Portfolio.objects.get(id=self.view.kwargs["portfolio_id"])
        if instrument_id := self.request.GET.get("portfolio_instrument"):
            instrument = Instrument.objects.get(id=instrument_id)
            return f"Composition between {str(portfolio)} and instrument portfolio {str(instrument)}"
        return f"Composition between {str(portfolio)} and related portfolios"



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class TransactionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_type", label="Type"),
                dp.Field(key="transaction_underlying_type", label="Underlying Type"),
                dp.Field(key="transaction_date", label="Transaction Date"),
                dp.Field(key="portfolio", label="Portfolio"),
                dp.Field(key="underlying_instrument", label="Instrument"),
                dp.Field(key="currency_fx_rate", label="FX Rate"),
                dp.Field(key="currency", label="Currency"),
                dp.Field(key="total_value", label="Total Value"),
                dp.Field(key="total_value_fx_portfolio", label="Total Value (Portfolio)"),
                dp.Field(key="total_value_usd", label="Total Value ($)"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["transaction_type", "underlying_instrument", "external_id"],
                ["transaction_date", "book_date", "value_date"],
                ["currency", "total_value", "total_value_fx_portfolio"],
                [".", "total_value_gross", "total_value_gross_fx_portfolio"],
                [repeat_field(3, "comment")],
            ]
        )


class TransactionPortfolioDisplayConfig(TransactionDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_type", label="Type"),
                dp.Field(key="transaction_underlying_type", label="Underlying Type"),
                dp.Field(key="transaction_date", label="Transaction Date"),
                dp.Field(key="underlying_instrument", label="Instrument"),
                dp.Field(key="currency_fx_rate", label="FX Rate"),
                dp.Field(key="currency", label="Currency"),
                dp.Field(key="total_value", label="Total Value"),
                dp.Field(key="total_value_fx_portfolio", label="Total Value (Portfolio)"),
                dp.Field(key="total_value_usd", label="Total Value ($)"),
            ]
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PortfolioInstrumentPreferredClassificationThroughEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-preferredclassification-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class InstrumentPortfolioThroughPortfolioModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-instrument-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AssetPositionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbportfolio:assetposition-list", request=self.request)


class AssetPositionPortfolioEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-asset-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class AssetPositionEquityEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:equity-asset-list",
            args=[self.view.kwargs["equity_id"]],
            request=self.request,
        )


class AssetPositionInstrumentEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-asset-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class AssetPositionIndexEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:index-asset-list",
            args=[self.view.kwargs["index_id"]],
            request=self.request,
        )


class AssetPositionProductGroupEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:productgroup-asset-list",
            args=[self.view.kwargs["product_group_id"]],
            request=self.request,
        )


class CashPositionPortfolioEndpointConfig(AssetPositionEndpointConfig):
    PK_FIELD = "portfolio"

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:productcashposition-list",
            args=[],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-list",
            args=[],
            request=self.request,
        )

    def get_update_endpoint(self, **kwargs):
        return None


class ContributorPortfolioChartEndpointConfig(AssetPositionEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-contributor-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class AssetPositionUnderlyingInstrumentChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:instrument-assetpositionchart-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class DistributionChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-distributionchart-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )


class DistributionTableEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-distributiontable-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )

    def get_endpoint(self, **kwargs):
        return None


class CompositionModelPortfolioPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-modelcompositionpandas-list",
            args=[self.view.kwargs["portfolio_id"]],
            request=self.request,
        )



from decimal import Decimal

from wbfdm.models import InstrumentPrice

from wbportfolio.pms.typing import Portfolio
from wbportfolio.rebalancing.base import AbstractRebalancingModel
from wbportfolio.rebalancing.decorators import register


@register("Equally Weighted Rebalancing")
class EquallyWeightedRebalancing(AbstractRebalancingModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.assets = self.portfolio.assets.filter(date=self.last_effective_date)

    def is_valid(self) -> bool:
        return (
            self.assets.exists()
            and InstrumentPrice.objects.filter(
                date=self.trade_date, instrument__in=self.assets.values("underlying_quote")
            ).exists()
        )

    def get_target_portfolio(self) -> Portfolio:
        positions = []
        assets = self.portfolio.assets.filter(date=self.last_effective_date)
        nb_assets = assets.count()
        for asset in assets:
            asset.date = self.trade_date
            asset.asset_valuation_date = self.trade_date
            positions.append(asset._build_dto(new_weight=Decimal(1 / nb_assets)))
        return Portfolio(positions=tuple(positions))



from decimal import Decimal

from django.db.models import (
    Case,
    DecimalField,
    ExpressionWrapper,
    F,
    OuterRef,
    Subquery,
    Sum,
    When,
)
from wbcore import viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.strings import format_number

from wbportfolio.filters import TransactionFilterSet, TransactionPortfolioFilterSet
from wbportfolio.models import Fees, Trade, Transaction
from wbportfolio.serializers import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)

from ..configs import (
    TransactionDisplayConfig,
    TransactionEndpointConfig,
    TransactionPortfolioDisplayConfig,
    TransactionPortfolioEndpointConfig,
    TransactionPortfolioTitleConfig,
)
from ..mixins import UserPortfolioRequestPermissionMixin


class TransactionRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:transaction"
    filterset_class = TransactionFilterSet
    search_fields = ("portfolio__name", "underlying_instrument__name", "comment")
    ordering_fields = (
        "transaction_date",
        "currency__key",
        "underlying_instrument__name",
        "portfolio__name",
        "transaction_type",
        "book_date",
        "value_date",
        "currency_fx_rate",
        "total_value",
        "total_value_fx_portfolio",
        "total_value_gross",
        "total_value_gross_fx_portfolio",
    )
    ordering = ["-transaction_date"]

    serializer_class = TransactionRepresentationSerializer
    queryset = Transaction.objects.all()


class TransactionModelViewSet(UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbportfolio:transaction"

    serializer_class = TransactionModelSerializer

    filterset_class = TransactionFilterSet
    ordering_fields = TransactionRepresentationViewSet.ordering_fields
    search_fields = TransactionRepresentationViewSet.search_fields
    ordering = TransactionRepresentationViewSet.ordering

    queryset = Transaction.objects.all()
    display_config_class = TransactionDisplayConfig
    endpoint_config_class = TransactionEndpointConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                fx_rate=CurrencyFXRates.get_fx_rates_subquery(
                    "transaction_date", currency="currency", lookup_expr="exact"
                ),
                total_value_usd=ExpressionWrapper(F("total_value"), output_field=DecimalField()),
                total_value_gross_usd=ExpressionWrapper(F("total_value_gross"), output_field=DecimalField()),
                transaction_underlying_type_trade=Subquery(
                    Trade.objects.filter(id=OuterRef("pk")).values("transaction_subtype")[:1]
                ),
                transaction_underlying_type_fees=Subquery(
                    Fees.objects.filter(id=OuterRef("pk")).values("transaction_subtype")[:1]
                ),
                transaction_underlying_type=Case(
                    When(
                        transaction_underlying_type_trade__isnull=False,
                        then=F("transaction_underlying_type_trade"),
                    ),
                    When(transaction_underlying_type_fees__isnull=False, then=F("transaction_underlying_type_fees")),
                ),
            )
            .select_related("underlying_instrument")
            .select_related("portfolio")
            .select_related("currency")
        )

    def get_aggregates(self, queryset, paginated_queryset):
        return {
            "total_value_usd": {
                "Σ": format_number(queryset.aggregate(s=Sum("total_value_usd"))["s"] or Decimal(0)),
            }
        }


class TransactionPortfolioModelViewSet(TransactionModelViewSet):
    IDENTIFIER = "wbportfolio:transaction"

    filterset_class = TransactionPortfolioFilterSet
    display_config_class = TransactionPortfolioDisplayConfig
    endpoint_config_class = TransactionPortfolioEndpointConfig
    title_config_class = TransactionPortfolioTitleConfig

    def get_queryset(self):
        qs = super().get_queryset().filter(portfolio=self.portfolio)
        if self.is_portfolio_manager:
            return qs
        return qs.exclude(transaction_type=Transaction.Type.FEES)



from typing import Optional

from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class CustodianDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name", width=Unit.PIXEL(200)),
                dp.Field(key="mapping", label="Mapping", width=Unit.PIXEL(200)),
            ],
            formatting=[],
            legends=[],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["name", "mapping"]])
